@@ -1,5 +1,5 @@
 /*
- * Copyright 2019-2023 OARC, Inc.
+ * Copyright 2019-2024 OARC, Inc.
  * Copyright 2017-2018 Akamai Technologies
  * Copyright 2006-2016 Nominum, Inc.
  * All rights reserved.
@@ -66,7 +66,7 @@
 
 #define TIMEOUT_CHECK_TIME 100000
 
-#define MAX_INPUT_DATA (64 * 1024)
+#define MAX_INPUT_DATA (64 * 1024) + 2
 
 #define MAX_SOCKETS 256
 
@@ -86,7 +86,7 @@ typedef struct {
     uint32_t            bufsize;
     bool                edns;
     bool                dnssec;
-    perf_tsigkey_t*     tsigkey;
+    const char*         tsigkey;
     perf_ednsoption_t*  edns_option;
     uint32_t            max_outstanding;
     uint32_t            max_qps;
@@ -129,7 +129,7 @@ typedef struct {
     uint64_t latency_min;
     uint64_t latency_max;
 
-    uint64_t num_conn_reconnect;
+    uint64_t num_conn_attempts;
     uint64_t num_conn_completed;
 
     uint64_t conn_latency_sum;
@@ -294,7 +294,7 @@ diff_stats(const config_t* config, stats_t* last, stats_t* now, stats_t* diff)
     diff->latency_min         = 0; /* not enough data */
     diff->latency_max         = 0;
 
-    diff->num_conn_reconnect = now->num_conn_reconnect - last->num_conn_reconnect;
+    diff->num_conn_attempts  = now->num_conn_attempts - last->num_conn_attempts;
     diff->num_conn_completed = now->num_conn_completed - last->num_conn_completed;
 
     diff->conn_latency_sum         = now->conn_latency_sum - last->conn_latency_sum;
@@ -438,16 +438,16 @@ print_statistics(const config_t* config, const times_t* times, stats_t* stats, u
 
     printf("\n");
 
-    if (!stats->num_conn_completed && !stats->num_conn_reconnect) {
+    if (!stats->num_conn_completed && !stats->num_conn_attempts) {
         fflush(stdout);
         return;
     }
 
     printf("Connection Statistics:\n\n");
-    printf("  Reconnections:        %" PRIu64 " (%.2lf%% of %" PRIu64 " connections)\n\n",
-        stats->num_conn_reconnect,
-        PERF_SAFE_DIV(100.0 * stats->num_conn_reconnect, stats->num_conn_completed),
-        stats->num_conn_completed);
+    printf("  Connection attempts:  %" PRIu64 " (%" PRIu64 " successful, %.2lf%%)\n\n",
+        stats->num_conn_attempts,
+        stats->num_conn_completed,
+        PERF_SAFE_DIV(100.0 * stats->num_conn_completed, stats->num_conn_attempts));
     latency_avg = PERF_SAFE_DIV(stats->conn_latency_sum, stats->num_conn_completed);
     printf("  Average Latency (s):  %u.%06u",
         (unsigned int)(latency_avg / MILLION),
@@ -518,7 +518,7 @@ sum_stats(const config_t* config, stats_t* total)
             total->latency_max = stats->latency_max;
 
         total->num_conn_completed += stats->num_conn_completed;
-        total->num_conn_reconnect += stats->num_conn_reconnect;
+        total->num_conn_attempts += stats->num_conn_attempts;
 
         total->conn_latency_sum += stats->conn_latency_sum;
         total->conn_latency_sum_squares += stats->conn_latency_sum_squares;
@@ -582,7 +582,6 @@ setup(int argc, char** argv, config_t* config)
     in_port_t   local_port     = DEFAULT_LOCAL_PORT;
     const char* filename       = NULL;
     const char* edns_option    = NULL;
-    const char* tsigkey        = NULL;
     const char* mode           = 0;
     const char* doh_uri        = DEFAULT_DOH_URI;
     const char* doh_method     = DEFAULT_DOH_METHOD;
@@ -646,7 +645,7 @@ setup(int argc, char** argv, config_t* config)
         &config->dnssec);
     perf_opt_add('y', perf_opt_string, "[alg:]name:secret",
         "the TSIG algorithm, name and secret (base64)", NULL,
-        &tsigkey);
+        &config->tsigkey);
     perf_opt_add('q', perf_opt_uint, "num_queries",
         "the maximum number of queries outstanding",
         stringify(DEFAULT_MAX_OUTSTANDING),
@@ -753,8 +752,11 @@ setup(int argc, char** argv, config_t* config)
     if (config->dnssec || edns_option != NULL)
         config->edns = true;
 
-    if (tsigkey != NULL)
-        config->tsigkey = perf_tsig_parsekey(tsigkey);
+    if (config->tsigkey) {
+        // check TSIG key to die earlier than in threads
+        perf_tsigkey_t* k = perf_tsig_parsekey(config->tsigkey);
+        perf_tsig_destroykey(&k);
+    }
 
     if (edns_option != NULL)
         config->edns_option = perf_edns_parseoption(edns_option);
@@ -802,8 +804,6 @@ cleanup(config_t* config)
         close(mainpipe[i]);
         close(intrpipe[i]);
     }
-    if (config->tsigkey != NULL)
-        perf_tsig_destroykey(&config->tsigkey);
     if (config->edns_option != NULL)
         perf_edns_destroyoption(&config->edns_option);
 }
@@ -853,6 +853,40 @@ wait_for_start(void)
     PERF_UNLOCK(&start_lock);
 }
 
+static inline void
+bit_set(unsigned char* bits, unsigned int bit)
+{
+    unsigned int shift, mask;
+
+    shift = 7 - (bit % 8);
+    mask  = 1 << shift;
+
+    bits[bit / 8] |= mask;
+}
+
+static inline void
+bit_clear(unsigned char* bits, unsigned int bit)
+{
+    unsigned int shift, mask;
+
+    shift = 7 - (bit % 8);
+    mask  = 1 << shift;
+
+    bits[bit / 8] &= ~mask;
+}
+
+static inline bool
+bit_check(unsigned char* bits, unsigned int bit)
+{
+    unsigned int shift;
+
+    shift = 7 - (bit % 8);
+
+    if ((bits[bit / 8] >> shift) & 0x01)
+        return true;
+    return false;
+}
+
 static void*
 do_send(void* arg)
 {
@@ -871,9 +905,11 @@ do_send(void* arg)
     unsigned char   packet_buffer[MAX_EDNS_PACKET];
     unsigned char*  base;
     unsigned int    length;
-    int             n, i, any_inprogress = 0;
+    int             n, i, any_inprogress = 0, sock = 0;
     perf_result_t   result;
     bool            all_fail;
+    unsigned char   socketbits[(MAX_SOCKETS / 8) + 1] = {};
+    perf_tsigkey_t* tsigkey                           = 0;
 
     tinfo           = (threadinfo_t*)arg;
     config          = tinfo->config;
@@ -882,6 +918,9 @@ do_send(void* arg)
     max_packet_size = config->edns ? MAX_EDNS_PACKET : MAX_UDP_PACKET;
     perf_buffer_init(&msg, packet_buffer, max_packet_size);
     perf_buffer_init(&lines, input_data, sizeof(input_data));
+    if (config->tsigkey) {
+        tsigkey = perf_tsig_parsekey(config->tsigkey);
+    }
 
     if (tinfo->max_qps > 0) {
         q_step = MILLION / tinfo->max_qps;
@@ -900,6 +939,21 @@ do_send(void* arg)
             now = perf_get_time();
         }
 
+        /* Some sock might still be sending, try flush all of them */
+        if (any_inprogress) {
+            any_inprogress = 0;
+            for (i = 0; i < tinfo->nsocks; i++) {
+                if (!bit_check(socketbits, i)) {
+                    continue;
+                }
+                if (!perf_net_sockready(tinfo->socks[i], threadpipe[0], TIMEOUT_CHECK_TIME)) {
+                    any_inprogress = 1;
+                } else {
+                    bit_clear(socketbits, i);
+                }
+            }
+        }
+
         /* Rate limiting */
         if (tinfo->max_qps > 0) {
             /* the 1 second time slice where q_sent is calculated over */
@@ -910,36 +964,40 @@ do_send(void* arg)
             }
             /* limit QPS over the 1 second slice */
             if (q_sent >= tinfo->max_qps) {
-                wait_us = q_slice - now;
-                if (config->qps_threshold_wait && wait_us > config->qps_threshold_wait) {
-                    wait_us -= config->qps_threshold_wait;
-                    struct timespec ts = { 0, 0 };
-                    if (wait_us >= MILLION) {
-                        ts.tv_sec  = wait_us / MILLION;
-                        ts.tv_nsec = (wait_us % MILLION) * 1000;
-                    } else {
-                        ts.tv_sec  = 0;
-                        ts.tv_nsec = wait_us * 1000;
+                if (!any_inprogress) { // only if nothing is in-progress
+                    wait_us = q_slice - now;
+                    if (config->qps_threshold_wait && wait_us > config->qps_threshold_wait) {
+                        wait_us -= config->qps_threshold_wait;
+                        struct timespec ts = { 0, 0 };
+                        if (wait_us >= MILLION) {
+                            ts.tv_sec  = wait_us / MILLION;
+                            ts.tv_nsec = (wait_us % MILLION) * 1000;
+                        } else {
+                            ts.tv_sec  = 0;
+                            ts.tv_nsec = wait_us * 1000;
+                        }
+                        nanosleep(&ts, NULL);
                     }
-                    nanosleep(&ts, NULL);
                 }
                 now = perf_get_time();
                 continue;
             }
             /* handle stepping to the next window to send a query on */
             if (req_time > now) {
-                wait_us = req_time - now;
-                if (config->qps_threshold_wait && wait_us > config->qps_threshold_wait) {
-                    wait_us -= config->qps_threshold_wait;
-                    struct timespec ts = { 0, 0 };
-                    if (wait_us >= MILLION) {
-                        ts.tv_sec  = wait_us / MILLION;
-                        ts.tv_nsec = (wait_us % MILLION) * 1000;
-                    } else {
-                        ts.tv_sec  = 0;
-                        ts.tv_nsec = wait_us * 1000;
+                if (!any_inprogress) { // only if nothing is in-progress
+                    wait_us = req_time - now;
+                    if (config->qps_threshold_wait && wait_us > config->qps_threshold_wait) {
+                        wait_us -= config->qps_threshold_wait;
+                        struct timespec ts = { 0, 0 };
+                        if (wait_us >= MILLION) {
+                            ts.tv_sec  = wait_us / MILLION;
+                            ts.tv_nsec = (wait_us % MILLION) * 1000;
+                        } else {
+                            ts.tv_sec  = 0;
+                            ts.tv_nsec = wait_us * 1000;
+                        }
+                        nanosleep(&ts, NULL);
                     }
-                    nanosleep(&ts, NULL);
                 }
                 now = perf_get_time();
                 continue;
@@ -951,7 +1009,9 @@ do_send(void* arg)
 
         /* Limit in-flight queries */
         if (num_outstanding(stats) >= tinfo->max_outstanding) {
-            PERF_TIMEDWAIT(&tinfo->cond, &tinfo->lock, &times->stop_time_ns, NULL);
+            if (!any_inprogress) { // only if nothing is in-progress
+                PERF_TIMEDWAIT(&tinfo->cond, &tinfo->lock, &times->stop_time_ns, NULL);
+            }
             PERF_UNLOCK(&tinfo->lock);
             now = perf_get_time();
             continue;
@@ -964,7 +1024,8 @@ do_send(void* arg)
         i        = tinfo->nsocks * 2;
         all_fail = true;
         while (i--) {
-            q->sock = tinfo->socks[tinfo->current_sock++ % tinfo->nsocks];
+            sock    = tinfo->current_sock++ % tinfo->nsocks;
+            q->sock = tinfo->socks[sock];
             switch (perf_net_sockready(q->sock, threadpipe[0], TIMEOUT_CHECK_TIME)) {
             case 0:
                 if (config->verbose && !config->suppress.sockready) {
@@ -1015,7 +1076,7 @@ do_send(void* arg)
             perf_buffer_usedregion(&lines, &used);
             result = perf_dns_buildrequest(&used, qid,
                 config->edns, config->dnssec, config->input_format == input_format_text_update,
-                config->tsigkey, config->edns_option,
+                tsigkey, config->edns_option,
                 &msg);
             break;
 
@@ -1060,6 +1121,7 @@ do_send(void* arg)
                     perf_log_warning("network congested, packet sending in progress");
                 }
                 any_inprogress = 1;
+                bit_set(socketbits, sock);
             } else {
                 if (config->verbose && !config->suppress.sendfailed) {
                     char __s[256];
@@ -1098,6 +1160,10 @@ do_send(void* arg)
     tinfo->done_sending   = true;
     if (write(mainpipe[1], "", 1)) { // lgtm [cpp/empty-block]
     }
+
+    if (tsigkey) {
+        perf_tsig_destroykey(&tsigkey);
+    }
     return NULL;
 }
 
@@ -1185,29 +1251,6 @@ recv_one(threadinfo_t* tinfo, int which_sock,
     return true;
 }
 
-static inline void
-bit_set(unsigned char* bits, unsigned int bit)
-{
-    unsigned int shift, mask;
-
-    shift = 7 - (bit % 8);
-    mask  = 1 << shift;
-
-    bits[bit / 8] |= mask;
-}
-
-static inline bool
-bit_check(unsigned char* bits, unsigned int bit)
-{
-    unsigned int shift;
-
-    shift = 7 - (bit % 8);
-
-    if ((bits[bit / 8] >> shift) & 0x01)
-        return true;
-    return false;
-}
-
 static void*
 do_recv(void* arg)
 {
@@ -1217,7 +1260,7 @@ do_recv(void* arg)
     received_query_t recvd[RECV_BATCH_SIZE] = { { 0, 0, 0, 0, 0, 0, false, false, 0 } };
     unsigned int     nrecvd;
     int              saved_errno;
-    unsigned char    socketbits[MAX_SOCKETS / 8];
+    unsigned char    socketbits[(MAX_SOCKETS / 8) + 1];
     uint64_t         now, latency;
     query_info*      q;
     unsigned int     current_socket, last_socket;
@@ -1467,7 +1510,8 @@ static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t ev
         break;
 
     case perf_socket_event_reconnecting:
-        stats->num_conn_reconnect++;
+    case perf_socket_event_connecting:
+        stats->num_conn_attempts++;
         break;
 
     default:
@@ -1477,7 +1521,7 @@ static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t ev
 
 static void
 threadinfo_init(threadinfo_t* tinfo, const config_t* config,
-    const times_t* times)
+    const times_t* times, int idx)
 {
     unsigned int offset, socket_offset, i;
 
@@ -1546,8 +1590,13 @@ threadinfo_init(threadinfo_t* tinfo, const config_t* config,
     }
     tinfo->current_sock = 0;
 
+    char name[16]; // glibc is limited to 16 characters
     PERF_THREAD(&tinfo->receiver, do_recv, tinfo);
+    snprintf(name, sizeof(name), "perf-recv-%04d", idx);
+    perf_os_thread_setname(tinfo->receiver, name);
     PERF_THREAD(&tinfo->sender, do_send, tinfo);
+    snprintf(name, sizeof(name), "perf-send-%04d", idx);
+    perf_os_thread_setname(tinfo->sender, name);
 }
 
 static void
@@ -1617,7 +1666,7 @@ int main(int argc, char** argv)
         perf_log_fatal("out of memory");
     }
     for (i = 0; i < config.threads; i++)
-        threadinfo_init(&threads[i], &config, &times);
+        threadinfo_init(&threads[i], &config, &times, i);
     if (config.stats_interval > 0) {
         stats_thread.config = &config;
         stats_thread.times  = &times;
@@ -1640,8 +1689,7 @@ int main(int argc, char** argv)
     perf_os_handlesignal(SIGINT, handle_sigint);
     perf_os_blocksignal(SIGINT, false);
     sock.fd = mainpipe[0];
-    result  = perf_os_waituntilreadable(&sock, intrpipe[0],
-        times.stop_time - times.start_time);
+    result  = perf_os_waituntilreadable(&sock, intrpipe[0], times.stop_time - times.start_time);
     if (result == PERF_R_CANCELED)
         interrupted = true;
 
