@@ -1,5 +1,5 @@
 /*
- * Copyright 2019-2022 OARC, Inc.
+ * Copyright 2019-2024 OARC, Inc.
  * Copyright 2017-2018 Akamai Technologies
  * Copyright 2006-2016 Nominum, Inc.
  * All rights reserved.
@@ -61,7 +61,7 @@
 #define DEFAULT_MAX_OUTSTANDING (64 * 1024)
 #define DEFAULT_MAX_FALL_BEHIND 1000
 
-#define MAX_INPUT_DATA (64 * 1024)
+#define MAX_INPUT_DATA (64 * 1024) + 2
 
 #define TIMEOUT_CHECK_TIME 5000000
 
@@ -149,7 +149,8 @@ static uint64_t num_queries_outstanding;
 static uint64_t num_responses_received;
 static uint64_t num_queries_timed_out;
 static uint64_t rcodecounts[16];
-static uint64_t num_reconnections;
+static uint64_t num_conn_completed;
+static uint64_t num_conn_attempts;
 
 static uint64_t time_now;
 static uint64_t time_of_program_start;
@@ -253,6 +254,7 @@ static void setup(int argc, char** argv)
     const char*  edns_option_str = NULL;
     const char*  doh_uri         = DEFAULT_DOH_URI;
     const char*  doh_method      = DEFAULT_DOH_METHOD;
+    const char*  tls_sni         = 0;
     const char*  local_suppress  = 0;
 
     size_t num_queries_per_conn = 0;
@@ -337,6 +339,8 @@ static void setup(int argc, char** argv)
         "the URI to use for DNS-over-HTTPS", DEFAULT_DOH_URI, &doh_uri);
     perf_long_opt_add("doh-method", perf_opt_string, "doh_method",
         "the HTTP method to use for DNS-over-HTTPS: GET or POST", DEFAULT_DOH_METHOD, &doh_method);
+    perf_long_opt_add("tls-sni", perf_opt_string, "tls_sni",
+        "the TLS SNI to use for TLS connections", NULL, &tls_sni);
     perf_long_opt_add("suppress", perf_opt_string, "message[,message,...]",
         "suppress messages/warnings, see dnsperf(1) man-page for list of message types", NULL, &local_suppress);
     perf_long_opt_add("num-queries-per-conn", perf_opt_uint, "queries",
@@ -367,6 +371,10 @@ static void setup(int argc, char** argv)
         }
     }
 
+    if (tls_sni) {
+        perf_net_tls_sni = tls_sni;
+    }
+
     if (doh_uri) {
         perf_net_doh_parse_uri(doh_uri);
     }
@@ -482,10 +490,12 @@ static void perf__net_event(struct perf_net_socket* sock, perf_socket_event_t ev
     case perf_socket_event_connected:
         b->connections++;
         b->conn_latency_sum += elapsed_time / (double)MILLION;
+        num_conn_completed++;
         break;
 
     case perf_socket_event_reconnecting:
-        num_reconnections++;
+    case perf_socket_event_connecting:
+        num_conn_attempts++;
         break;
 
     default:
@@ -538,8 +548,7 @@ print_statistics(void)
             perf_dns_rcode_strings[i], rcodecounts[i],
             (rcodecounts[i] * 100.0) / num_responses_received);
     }
-    printf("\n");
-    printf("  Reconnection(s):      %" PRIu64 "\n", num_reconnections);
+    printf("\n\n");
     printf("  Run time (s):         %u.%06u\n",
         (unsigned int)(run_time / MILLION),
         (unsigned int)(run_time % MILLION));
@@ -562,6 +571,10 @@ print_statistics(void)
     printf("  Maximum throughput:   %.6lf qps\n", max_throughput);
     printf("  Lost at that point:   %.2f%%\n", loss_at_max_throughput);
     printf("\n");
+    printf("  Connection attempts:  %" PRIu64 " (%" PRIu64 " successful, %.2lf%%)\n\n",
+        num_conn_attempts,
+        num_conn_completed,
+        PERF_SAFE_DIV(100.0 * num_conn_completed, num_conn_attempts));
 }
 
 /*
